@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP, TypeSynonymInstances #-}
+{-# LANGUAGE CPP, TypeSynonymInstances, FlexibleInstances #-}
 {-# OPTIONS -fno-cse #-}
 
 module Agda.Interaction.GhciTop
@@ -42,7 +42,6 @@ import Agda.Utils.String
 import Agda.Utils.FileName
 import qualified Agda.Utils.Trie as Trie
 import Agda.Utils.Tuple
-import qualified Agda.Utils.IO.Locale as LocIO
 import qualified Agda.Utils.IO.UTF8 as UTF8
 
 import Control.Monad.Error
@@ -63,8 +62,8 @@ import qualified Agda.TypeChecking.Monad as TM
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Errors
-import Agda.TypeChecking.Serialise (encodeFile)
 
+import Agda.Syntax.Fixity
 import Agda.Syntax.Position
 import Agda.Syntax.Parser
 import qualified Agda.Syntax.Parser.Tokens as T
@@ -83,12 +82,12 @@ import Agda.Syntax.Translation.AbstractToConcrete hiding (withScope)
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Abstract.Name
 
+import Agda.Interaction.EmacsCommand
 import Agda.Interaction.Exceptions
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import Agda.Interaction.MakeCase
 import qualified Agda.Interaction.BasicOps as B
-import qualified Agda.Interaction.CommandLine.CommandLine as CL
 import Agda.Interaction.Highlighting.Emacs
 import Agda.Interaction.Highlighting.Generate
 import qualified Agda.Interaction.Imports as Imp
@@ -97,6 +96,7 @@ import Agda.Termination.TermCheck
 
 import qualified Agda.Compiler.Epic.Compiler as Epic
 import qualified Agda.Compiler.MAlonzo.Compiler as MAlonzo
+import qualified Agda.Compiler.JS.Compiler as JS
 
 import qualified Agda.Auto.Auto as Auto
 
@@ -173,7 +173,7 @@ ioTCM_ m = do
     Left err -> do
       Right doc <- runTCM $ prettyTCM err
       putStrLn $ render doc
-      return $ undefined
+      return __IMPOSSIBLE__
 {-
   Right (x, s) <- runTCM $ do
     put $ theTCState tcs
@@ -205,9 +205,7 @@ ioTCM current highlightingFile cmd = infoOnException $ do
   current <- absolute current
 
   -- Read the state.
-  State { theTCState     = st
-        , theCurrentFile = f
-        } <- readIORef theState
+  State { theTCState = st } <- readIORef theState
 
   -- Run the computation.
   r <- runTCM $ catchError (do
@@ -227,20 +225,31 @@ ioTCM current highlightingFile cmd = infoOnException $ do
            st <- get
            return (Right (x, st))
          ) (\e -> do
-           s <- prettyError e
-           return (Left (s, e))
+           pers <- stPersistent <$> get
+           s    <- prettyError e
+           return (Left (pers, s, e))
          )
 
-  -- Update the state.
+  -- Upon success: update the state. Upon failure: update the
+  -- persistent state, and, for independent commands, the current
+  -- file.
   case r of
     Right (Right (m, st')) ->
       modifyIORef theState $ \s ->
-        s { theTCState   = st'
+        s { theTCState = st'
           }
-    _ | isIndependent cmd ->
+    Right (Left (pers, _, _)) -> do
       modifyIORef theState $ \s ->
-        s { theCurrentFile = Nothing }
-    _ -> return ()
+        s { theTCState = (theTCState s) { stPersistent = pers }
+          }
+    Left _ -> return ()
+  when (isIndependent cmd) $
+    case r of
+      Right (Right _) -> return ()
+      _               ->
+        modifyIORef theState $ \s ->
+          s { theCurrentFile = Nothing
+            }
 
   -- Write out syntax highlighting info.
   case highlightingFile of
@@ -260,8 +269,8 @@ ioTCM current highlightingFile cmd = infoOnException $ do
               return ( iHighlighting $ miInterface mi
                      , stModuleToSource st'
                      )
-            Right (Left (s, e)) -> errHi e (Just s)
-            Left e              -> errHi e Nothing
+            Right (Left (_ , s, e)) -> errHi e (Just s)
+            Left e                  -> errHi e Nothing
 
   -- If an error was encountered, display an error message and exit
   -- with an error code; otherwise, inform Emacs about the buffer's
@@ -271,15 +280,15 @@ ioTCM current highlightingFile cmd = infoOnException $ do
                              optShowImplicit $ stPragmaOptions st
                          }
   case r of
-    Right (Left (s, e)) -> displayErrorAndExit errStatus (getRange e) s
-    Left e              -> displayErrorAndExit errStatus (getRange e) $
-                             tcErrString e
-    Right (Right _)     -> do
+    Right (Left (_, s, e)) -> displayErrorAndExit errStatus (getRange e) s
+    Left e                 -> displayErrorAndExit errStatus (getRange e) $
+                                tcErrString e
+    Right (Right _)        -> do
       f <- theCurrentFile <$> readIORef theState
       case f of
         Just (f, _) | f === current -> do
           is <- theInteractionPoints <$> liftIO (readIORef theState)
-          liftIO $ LocIO.putStrLn $ response $
+          liftIO $ putResponse $
             L [A "agda2-goals-action", Q $ L $ List.map showNumIId is]
         _ -> return ()
 
@@ -315,7 +324,7 @@ cmd_load' file includes unsolvedOK cmd =
     -- All options (except for the verbosity setting) are reset when a
     -- file is reloaded, including the choice of whether or not to
     -- display implicit arguments. (At this point the include
-    -- directories have already been reset, so they are preserved.)
+    -- directories have already been set, so they are preserved.)
     opts <- commandLineOptions
     setCommandLineOptions $
       defaultOptions { optIncludeDirs   = optIncludeDirs opts
@@ -327,9 +336,9 @@ cmd_load' file includes unsolvedOK cmd =
                      }
 
     -- Reset the state, preserving options and decoded modules. Note
-    -- that Imp.typeCheck resets the decoded modules if the include
-    -- directories have changed.
-    preserveDecodedModules resetState
+    -- that if the include directories have changed, then the decoded
+    -- modules are reset when cmd_load' is run by ioTCM.
+    resetState
 
     ok <- Imp.typeCheck f
 
@@ -352,7 +361,7 @@ cmd_load' file includes unsolvedOK cmd =
 
 -- | Available backends.
 
-data Backend = MAlonzo | Epic
+data Backend = MAlonzo | Epic | JS
 
 -- | @cmd_compile b m includes@ compiles the module in file @m@ using
 -- the backend @b@, using @includes@ as the include directories.
@@ -365,6 +374,7 @@ cmd_compile b file includes =
         case b of
           MAlonzo -> MAlonzo.compilerMain i
           Epic    -> Epic.compilerMain i
+          JS      -> JS.compilerMain i
         display_info "*Compilation result*"
                      "The module was successfully compiled."
       Just w ->
@@ -388,7 +398,7 @@ cmd_metas = Interaction Dependent $ do -- CL.showMetas []
   hms <- B.typesOfHiddenMetas B.Normalised
   if not $ null ims && null hms
     then do
-      di <- mapM (\i -> B.withInteractionId (B.outputFormId i) (showATop i)) ims
+      di <- mapM (\i -> B.withInteractionId (B.outputFormId $ B.OutputForm 0 i) (showATop i)) ims
       dh <- mapM showA' hms
       display_info "*All Goals*" $ unlines $ di ++ dh
       return Nothing
@@ -405,7 +415,7 @@ cmd_metas = Interaction Dependent $ do -- CL.showMetas []
     metaId _ = __IMPOSSIBLE__
     showA' m = do
       r <- getMetaRange (metaId m)
-      d <- B.withMetaId (B.outputFormId m) (showATop m)
+      d <- B.withMetaId (B.outputFormId $ B.OutputForm 0 m) (showATop m)
       return $ d ++ "  [ at " ++ show r ++ " ]"
 
 -- | If the range is 'noRange', then the string comes from the
@@ -434,8 +444,7 @@ give_gen' give_ref mk_newtxt ii rng s = do
   liftIO $ modifyIORef theState $ \s ->
              s { theInteractionPoints =
                    replace ii iis (theInteractionPoints s) }
-  liftIO $ LocIO.putStrLn $ response $
-             L [A "agda2-give-action", showNumIId ii, newtxt]
+  liftIO $ putResponse $ L [A "agda2-give-action", showNumIId ii, newtxt]
   command cmd_metas
   return Nothing
   where
@@ -472,7 +481,7 @@ cmd_auto ii rng s = Interaction Dependent $ do
     mapM_ (\(ii, s) -> do
       liftIO $ modifyIORef theState $ \s ->
         s { theInteractionPoints = filter (/= ii) (theInteractionPoints s) }
-      liftIO $ LocIO.putStrLn $ response $ L [A "agda2-give-action", showNumIId ii, A $ quote s]
+      liftIO $ putResponse $ L [A "agda2-give-action", showNumIId ii, A $ quote s]
      ) xs
     case msg of
      Nothing -> command cmd_metas >> return ()
@@ -482,7 +491,7 @@ cmd_auto ii rng s = Interaction Dependent $ do
     case msg of
      Nothing -> return ()
      Just msg -> display_info "*Auto*" msg
-    liftIO $ LocIO.putStrLn $ response $
+    liftIO $ putResponse $
      Cons (A "last")
           (L [ A "agda2-make-case-action",
                Q $ L $ List.map (A . quote) cs
@@ -518,7 +527,7 @@ prettyContext norm rev ii = B.withInteractionId ii $ do
   es  <- mapM (prettyATop . B.ofExpr) ctx
   ns  <- mapM (showATop   . B.ofName) ctx
   let shuffle = if rev then reverse else id
-  return $ align 10 $ shuffle $ zip ns (map (text ":" <+>) es)
+  return $ align 10 $ filter (not . null. fst) $ shuffle $ zip ns (map (text ":" <+>) es)
 
 cmd_context :: B.Rewrite -> GoalCommand
 cmd_context norm ii _ _ = Interaction Dependent $ do
@@ -655,15 +664,14 @@ showStatus s = intercalate "," $ catMaybes [checked, showImpl]
 
 displayStatus :: Status -> IO ()
 displayStatus s =
-  LocIO.putStrLn $ response $
-    L [A "agda2-status-action", A (quote $ showStatus s)]
+  putResponse $ L [A "agda2-status-action", A (quote $ showStatus s)]
 
 -- | @display_info' header content@ displays @content@ (with header
 -- @header@) in some suitable way.
 
 display_info' :: String -> String -> IO ()
 display_info' bufname content =
-  LocIO.putStrLn $ response $
+  putResponse $
     L [ A "agda2-info-action"
       , A (quote bufname)
       , A (quote content)
@@ -683,23 +691,6 @@ display_info bufname content = do
 display_infoD :: String -> Doc -> TCM ()
 display_infoD bufname content = display_info bufname (render content)
 
--- | Formats a response command.
-
-response :: Lisp String -> String
-response l = show (text "agda2_mode_code" <+> pretty l)
-
-data Lisp a = A a | L [Lisp a] | Q (Lisp a) | Cons (Lisp a) (Lisp a)
-
-instance Pretty a => Pretty (Lisp a) where
-  pretty (A a )     = pretty a
-  pretty (L xs)     = parens (sep (List.map pretty xs))
-  pretty (Q x)      = text "'" <> pretty x
-  pretty (Cons a b) = parens (pretty a <+> text "." <+> pretty b)
-
-instance Pretty String where pretty = text
-
-instance Pretty a => Show (Lisp a) where show = show . pretty
-
 showNumIId = A . tail . show
 
 takenNameStr :: TCM [String]
@@ -722,21 +713,40 @@ nameModifiers = "" : "'" : "''" : [show i | i <-[3..]]
 
 cmd_make_case :: GoalCommand
 cmd_make_case ii rng s = Interaction Dependent $ do
-  cs <- makeCase ii rng s
+  (casectxt , cs) <- makeCase ii rng s
   B.withInteractionId ii $ do
-    pcs <- mapM prettyA cs
-    liftIO $ LocIO.putStrLn $ response $
+    hidden <- showImplicitArguments
+    pcs <- mapM prettyA $ List.map (extlam_dropLLifted casectxt hidden) cs
+    liftIO $ putResponse $
       Cons (A "last")
-           (L [ A "agda2-make-case-action"
-              , Q $ L $ List.map (A . quote . render) pcs
+           (L [ A (emacscmd casectxt)
+              , Q $ L $ List.map (A . quote . (extlam_dropName casectxt) . render) pcs
               ])
   return Nothing
-  where render = renderStyle (style { mode = OneLineMode })
+  where
+    render = renderStyle (style { mode = OneLineMode })
+    emacscmd :: CaseContext -> String
+    emacscmd FunctionDef = "agda2-make-case-action"
+    emacscmd (ExtendedLambda _ _) = "agda2-make-case-action-extendlam"
+
+    -- Drops pattern added to extended lambda functions when lambda lifting them
+    extlam_dropLLifted :: CaseContext -> Bool -> SA.Clause -> SA.Clause
+    extlam_dropLLifted FunctionDef _ x = x
+    extlam_dropLLifted (ExtendedLambda h nh) hidden (SA.Clause (SA.LHS info name nps ps) rhs decl)
+      = let n = if hidden then h else nh
+        in
+         (SA.Clause (SA.LHS info name (drop n nps) ps) rhs decl)
+
+    -- very dirty hack, string manipulation by dropping the function name
+    -- and replacing " = " with " -> "
+    extlam_dropName :: CaseContext -> String -> String
+    extlam_dropName FunctionDef x = x
+    extlam_dropName (ExtendedLambda _ _) x = unwords $ map (\ y -> if y == "=" then "→" else y) $ drop 1 $ words x
 
 cmd_solveAll :: Interaction
 cmd_solveAll = Interaction Dependent $ do
   out <- getInsts
-  liftIO $ LocIO.putStrLn $ response $
+  liftIO $ putResponse $
     Cons (A "last")
          (L [ A "agda2-solveAll-action"
             , Q . L $ concatMap prn out
@@ -747,7 +757,7 @@ cmd_solveAll = Interaction Dependent $ do
     where
       lowr (i, m, e) = do
         mi <- getMetaInfo <$> lookupMeta m
-        e <- withMetaInfo mi $ lowerMeta <$> abstractToConcrete_ e
+        e <- withMetaInfo mi $ lowerMeta <$> abstractToConcreteCtx TopCtx e
         return (i, e)
   prn (ii,e)= [showNumIId ii, A $ quote $ show e]
 
@@ -766,6 +776,7 @@ instance LowerMeta SC.Expr where
       SC.WithApp r e es	   -> SC.WithApp r (lowerMeta e) (lowerMeta es)
       SC.Lam r bs e1       -> SC.Lam r (lowerMeta bs) (go e1)
       SC.AbsurdLam r h     -> SC.AbsurdLam r h
+      SC.ExtendedLam r cs  -> SC.ExtendedLam r cs
       SC.Fun r ae1 e2      -> SC.Fun r (lowerMeta ae1) (go e2)
       SC.Pi tb e1          -> SC.Pi (lowerMeta tb) (go e1)
       SC.Set _             -> e
@@ -782,12 +793,22 @@ instance LowerMeta SC.Expr where
       SC.RawApp r es	-> SC.RawApp r (lowerMeta es)
       SC.OpApp r x es	-> SC.OpApp r x (lowerMeta es)
       SC.Rec r fs	-> SC.Rec r (List.map (id -*- lowerMeta) fs)
+      SC.RecUpdate r e fs -> SC.RecUpdate r (lowerMeta e) (List.map (id -*- lowerMeta) fs)
       SC.HiddenArg r e	-> SC.HiddenArg r (lowerMeta e)
+      SC.InstanceArg r e  -> SC.InstanceArg r (lowerMeta e)
       SC.QuoteGoal r x e  -> SC.QuoteGoal r x (lowerMeta e)
-      SC.Quote r        -> SC.Quote r
+      e@SC.Quote{}      -> e
+      e@SC.QuoteTerm{}  -> e
+      e@SC.Unquote{}    -> e
+      SC.DontCare e     -> SC.DontCare (lowerMeta e)
+
+instance LowerMeta (OpApp SC.Expr) where
+  lowerMeta (Ordinary e) = Ordinary $ lowerMeta e
+  lowerMeta (SyntaxBindingLambda r bs e) = SyntaxBindingLambda r (lowerMeta bs) (lowerMeta e)
+
 
 instance LowerMeta SC.LamBinding where
-  lowerMeta b@(SC.DomainFree _ _) = b
+  lowerMeta b@(SC.DomainFree _ _ _) = b
   lowerMeta (SC.DomainFull tb)    = SC.DomainFull (lowerMeta tb)
 
 instance LowerMeta SC.TypedBindings where
@@ -801,19 +822,27 @@ instance LowerMeta SC.RHS where
     lowerMeta (SC.RHS e)    = SC.RHS (lowerMeta e)
     lowerMeta  SC.AbsurdRHS = SC.AbsurdRHS
 
+instance LowerMeta (Maybe SC.Expr) where
+    lowerMeta (Just e) = Just (lowerMeta e)
+    lowerMeta Nothing  = Nothing
+
 instance LowerMeta SC.Declaration where
   lowerMeta = go where
     go d = case d of
       TypeSig rel n e1              -> TypeSig rel n (lowerMeta e1)
       SC.Field n e1                 -> SC.Field n (lowerMeta e1)
       FunClause lhs rhs whcl        -> FunClause lhs (lowerMeta rhs) (lowerMeta whcl)
+      SC.DataSig r ind n tel e1     -> SC.DataSig r ind n
+                                         (lowerMeta tel) (lowerMeta e1)
       Data r ind n tel e1 cs        -> Data r ind n
                                          (lowerMeta tel) (lowerMeta e1) (lowerMeta cs)
+      SC.RecordSig r n tel e1       -> SC.RecordSig r n
+                                         (lowerMeta tel) (lowerMeta e1)
       SC.Record r n c tel e1 cs     -> SC.Record r n c
                                          (lowerMeta tel) (lowerMeta e1) (lowerMeta cs)
       Infix _ _                     -> d
       Syntax _ _                     -> d
-      Mutual r ds                   -> Mutual r (lowerMeta ds)
+      SC.Mutual r ds                -> SC.Mutual r (lowerMeta ds)
       Abstract r ds                 -> Abstract r (lowerMeta ds)
       Private r ds                  -> Private r (lowerMeta ds)
       Postulate r sigs              -> Postulate r (lowerMeta sigs)
@@ -821,10 +850,14 @@ instance LowerMeta SC.Declaration where
       SC.Open _ _ _                 -> d
       SC.Import _ _ _ _ _           -> d
       SC.Pragma _                   -> d
-      ModuleMacro r n tel e1 op dir -> ModuleMacro r n
-                                         (lowerMeta tel) (lowerMeta e1) op dir
+      ModuleMacro r n modapp op dir -> ModuleMacro r n
+                                         (lowerMeta modapp) op dir
       SC.Module r qn tel ds         -> SC.Module r qn (lowerMeta tel) (lowerMeta ds)
 
+instance LowerMeta SC.ModuleApplication where
+  lowerMeta (SC.SectionApp r tel e) = SC.SectionApp r (lowerMeta tel) (lowerMeta e)
+  lowerMeta (SC.RecordModuleIFS r rec) = SC.RecordModuleIFS r rec
+
 instance LowerMeta SC.WhereClause where
   lowerMeta SC.NoWhere		= SC.NoWhere
   lowerMeta (SC.AnyWhere ds)	= SC.AnyWhere $ lowerMeta ds
@@ -944,7 +977,7 @@ tellEmacsToJumpToError r = do
     Nothing                                    -> return ()
     Just (Pn { srcFile = Nothing })            -> return ()
     Just (Pn { srcFile = Just f, posPos = p }) ->
-      LocIO.putStrLn $ response $
+      putResponse $
         L [ A "annotation-goto"
           , Q $ L [A (quote $ filePath f), A ".", A (show p)]
           ]
