@@ -1,3 +1,5 @@
+{-# LANGUAGE StandaloneDeriving, DeriveFunctor #-}
+
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
     , PragmaOptions(..)
@@ -11,6 +13,7 @@ module Agda.Interaction.Options
     , defaultOptions
     , defaultVerbosity
     , standardOptions_
+    , unsafePragmaOptions
     , isLiterate
     , mapFlag
     , usage
@@ -19,7 +22,8 @@ module Agda.Interaction.Options
 
 import Control.Monad            ( when )
 import Control.Monad.Error	( MonadError(..) )
-import Data.List		( isSuffixOf )
+import Data.Maybe (isJust)
+import Data.List		( isSuffixOf , intercalate )
 import System.Console.GetOpt	(getOpt, usageInfo, ArgOrder(ReturnInOrder)
 				, OptDescr(..), ArgDescr(..)
 				)
@@ -38,13 +42,8 @@ isLiterate file = ".lagda" `isSuffixOf` file
 
 -- OptDescr is a Functor --------------------------------------------------
 
-instance Functor OptDescr where
-    fmap f (Option short long arg descr) = Option short long (fmap f arg) descr
-
-instance Functor ArgDescr where
-    fmap f (NoArg x)	= NoArg (f x)
-    fmap f (ReqArg p s) = ReqArg (f . p) s
-    fmap f (OptArg p s) = OptArg (f . p) s
+deriving instance Functor OptDescr
+deriving instance Functor ArgDescr
 
 type Verbosity = Trie String Int
 
@@ -62,10 +61,12 @@ data CommandLineOptions =
 	    , optRunTests             :: Bool
 	    , optCompile              :: Bool
 	    , optEpicCompile          :: Bool
+	    , optJSCompile            :: Bool
             , optCompileDir           :: Maybe FilePath
               -- ^ In the absence of a path the project root is used.
 	    , optGenerateVimFile      :: Bool
 	    , optGenerateHTML         :: Bool
+	    , optDependencyGraph      :: Maybe FilePath
 	    , optHTMLDir              :: FilePath
 	    , optCSSFile              :: Maybe FilePath
 	    , optIgnoreInterfaces     :: Bool
@@ -73,7 +74,8 @@ data CommandLineOptions =
             , optGhcFlags             :: [String]
             , optPragmaOptions        :: PragmaOptions
             , optEpicFlags            :: [String]
-	    }
+            , optSafe                 :: Bool
+            }
     deriving Show
 
 -- | Options which can be set in a pragma.
@@ -87,13 +89,13 @@ data PragmaOptions = PragmaOptions
   , optTerminationCheck          :: Bool
   , optTerminationDepth          :: Int
   , optCompletenessCheck         :: Bool
-  , optUnreachableCheck          :: Bool
   , optUniverseCheck             :: Bool
   , optSizedTypes                :: Bool
   , optInjectiveTypeConstructors :: Bool
   , optGuardingTypeConstructors  :: Bool
   , optUniversePolymorphism      :: Bool
   , optIrrelevantProjections     :: Bool
+  , optExperimentalIrrelevance   :: Bool  -- ^ irrelevant levels, irrelevant data matching
   , optWithoutK                  :: Bool
   }
   deriving Show
@@ -124,9 +126,11 @@ defaultOptions =
 	    , optRunTests             = False
 	    , optCompile              = False
 	    , optEpicCompile          = False
+	    , optJSCompile            = False
             , optCompileDir           = Nothing
 	    , optGenerateVimFile      = False
 	    , optGenerateHTML         = False
+	    , optDependencyGraph      = Nothing
 	    , optHTMLDir              = defaultHTMLDir
 	    , optCSSFile              = Nothing
 	    , optIgnoreInterfaces     = False
@@ -134,6 +138,7 @@ defaultOptions =
             , optGhcFlags             = []
             , optPragmaOptions        = defaultPragmaOptions
             , optEpicFlags            = []
+            , optSafe                 = False
 	    }
 
 defaultPragmaOptions :: PragmaOptions
@@ -141,18 +146,18 @@ defaultPragmaOptions = PragmaOptions
   { optShowImplicit              = False
   , optVerbose                   = defaultVerbosity
   , optProofIrrelevance          = False
+  , optExperimentalIrrelevance   = False
   , optIrrelevantProjections     = True
   , optAllowUnsolved             = False
   , optDisablePositivity         = False
   , optTerminationCheck          = True
   , optTerminationDepth          = 0    -- this is the cutoff value
   , optCompletenessCheck         = True
-  , optUnreachableCheck          = True
   , optUniverseCheck             = True
   , optSizedTypes                = False
   , optInjectiveTypeConstructors = False
   , optGuardingTypeConstructors  = False
-  , optUniversePolymorphism      = False
+  , optUniversePolymorphism      = True
   , optWithoutK                  = False
   }
 
@@ -176,10 +181,12 @@ checkOpts :: Flag CommandLineOptions
 checkOpts opts
   | not (atMostOne [optAllowUnsolved . p, optCompile]) = Left
       "Unsolved meta variables are not allowed when compiling.\n"
-  | not (atMostOne [optInteractive, optCompile, optEpicCompile]) =
+  | not (atMostOne [optInteractive, optCompile, optEpicCompile, optJSCompile]) =
       Left "Choose at most one: compiler/interactive mode.\n"
   | not (atMostOne [optGenerateHTML, optInteractive]) =
       Left "Choose at most one: HTML generator or interactive mode.\n"
+  | not (atMostOne [isJust . optDependencyGraph, optInteractive]) =
+      Left "Choose at most one: Dependency graph generator or interactive mode.\n"
   | not (atMostOne [ optUniversePolymorphism . p
                    , not . optUniverseCheck . p
                    ]) =
@@ -193,6 +200,30 @@ checkOpts opts
 
   p = optPragmaOptions
 
+-- Check for unsafe pramas. Gives a list of used unsafe flags.
+
+unsafePragmaOptions :: PragmaOptions -> [String]
+unsafePragmaOptions opts =
+  [ "--allow-unsolved-metas"                     | optAllowUnsolved opts             ] ++
+  [ "--no-positivity-check"                      | optDisablePositivity opts         ] ++
+  [ "--no-termination-check"                     | not (optTerminationCheck opts)    ] ++
+  [ "--no-coverage-check"                        | not (optCompletenessCheck opts)   ] ++
+  [ "--type-in-type"                             | not (optUniverseCheck opts)       ] ++
+  [ "--sized-types"                              | optSizedTypes opts                ] ++
+  [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
+  [ "--guardedness-preserving-type-constructors" | optGuardingTypeConstructors opts  ] ++
+  [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ]
+
+-- The default pragma options should be considered safe
+
+defaultPragmaOptionsSafe :: IO Bool
+defaultPragmaOptionsSafe
+    | null unsafe = return True
+    | otherwise   = do putStrLn $ "Following pragmas are default but not safe: "
+                                  ++ intercalate ", " unsafe
+                       return False
+  where unsafe = unsafePragmaOptions defaultPragmaOptions
+
 inputFlag :: FilePath -> Flag CommandLineOptions
 inputFlag f o =
     case optInputFile o of
@@ -201,7 +232,9 @@ inputFlag f o =
 
 versionFlag                  o = return $ o { optShowVersion               = True  }
 helpFlag                     o = return $ o { optShowHelp                  = True  }
+safeFlag                     o = return $ o { optSafe                      = True  }
 proofIrrelevanceFlag         o = return $ o { optProofIrrelevance          = True  }
+experimentalIrrelevanceFlag  o = return $ o { optExperimentalIrrelevance   = True  }
 noIrrelevantProjectionsFlag  o = return $ o { optIrrelevantProjections     = False }
 ignoreInterfacesFlag         o = return $ o { optIgnoreInterfaces          = True  }
 allowUnsolvedFlag            o = return $ o { optAllowUnsolved             = True  }
@@ -211,12 +244,13 @@ vimFlag                      o = return $ o { optGenerateVimFile           = Tru
 noPositivityFlag             o = return $ o { optDisablePositivity         = True  }
 dontTerminationCheckFlag     o = return $ o { optTerminationCheck          = False }
 dontCompletenessCheckFlag    o = return $ o { optCompletenessCheck         = False }
-noUnreachableCheckFlag       o = return $ o { optUnreachableCheck          = False }
-dontUniverseCheckFlag        o = return $ o { optUniverseCheck             = False }
+dontUniverseCheckFlag        o = return $ o { optUniverseCheck             = False
+                                            , optUniversePolymorphism      = False }
 sizedTypes                   o = return $ o { optSizedTypes                = True  }
 injectiveTypeConstructorFlag o = return $ o { optInjectiveTypeConstructors = True  }
 guardingTypeConstructorFlag  o = return $ o { optGuardingTypeConstructors  = True  }
 universePolymorphismFlag     o = return $ o { optUniversePolymorphism      = True  }
+noUniversePolymorphismFlag   o = return $ o { optUniversePolymorphism      = False }
 noForcingFlag                o = return $ o { optForcing                   = False }
 withoutKFlag                 o = return $ o { optWithoutK                  = True  }
 
@@ -226,11 +260,13 @@ interactiveFlag  o = return $ o { optInteractive    = True
 		                }
 compileFlag      o = return $ o { optCompile    = True }
 compileEpicFlag  o = return $ o { optEpicCompile = True}
+compileJSFlag    o = return $ o { optJSCompile = True}
 compileDirFlag f o = return $ o { optCompileDir = Just f }
 ghcFlag        f o = return $ o { optGhcFlags   = f : optGhcFlags o }
 epicFlagsFlag  s o = return $ o { optEpicFlags  = optEpicFlags o ++ [s]}
 
 htmlFlag      o = return $ o { optGenerateHTML = True }
+dependencyGraphFlag f o = return $ o { optDependencyGraph  = Just f }
 htmlDirFlag d o = return $ o { optHTMLDir      = d }
 cssFlag     f o = return $ o { optCSSFile      = Just f }
 
@@ -268,6 +304,7 @@ standardOptions =
     , Option ['c']  ["compile"] (NoArg compileFlag)
                     "compile program using the MAlonzo backend (experimental)"
     , Option []     ["epic"] (NoArg compileEpicFlag) "compile program using the Epic backend"
+    , Option []     ["js"] (NoArg compileJSFlag) "compile program using the JS backend"
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
 		    ("directory for compiler output (default: the project root)")
     , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
@@ -280,6 +317,8 @@ standardOptions =
 		    "generate Vim highlighting files"
     , Option []	    ["html"] (NoArg htmlFlag)
 		    "generate HTML files with highlighted source code"
+    , Option []	    ["dependency-graph"] (ReqArg dependencyGraphFlag "FILE")
+		    "generate a Dot file with a module dependency graph"
     , Option []	    ["html-dir"] (ReqArg htmlDirFlag "DIR")
                     ("directory in which HTML files are placed (default: " ++
                      defaultHTMLDir ++ ")")
@@ -291,6 +330,8 @@ standardOptions =
 		    "look for imports in DIR"
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
                     "disable the forcing optimisation"
+    , Option []     ["safe"] (NoArg safeFlag)
+                    "disable postulates, unsafe OPTION pragmas and primTrustMe"
     ] ++ map (fmap lift) pragmaOptions
   where
   lift :: Flag PragmaOptions -> Flag CommandLineOptions
@@ -316,20 +357,22 @@ pragmaOptions =
 		    "allow termination checker to count decrease/increase upto N (default N=1)"
     , Option []	    ["no-coverage-check"] (NoArg dontCompletenessCheckFlag)
 		    "do not warn about possibly incomplete pattern matches"
-    , Option []	    ["no-unreachable-check"] (NoArg noUnreachableCheckFlag)
-		    "do not warn about unreachable function clauses"
     , Option []	    ["type-in-type"] (NoArg dontUniverseCheckFlag)
 		    "ignore universe levels (this makes Agda inconsistent)"
     , Option []     ["sized-types"] (NoArg sizedTypes)
-                    "use sized datatypes"
+                    "use sized types (inconsistent with coinduction)"
     , Option []     ["injective-type-constructors"] (NoArg injectiveTypeConstructorFlag)
                     "enable injective type constructors (makes Agda anti-classical and possibly inconsistent)"
     , Option []     ["guardedness-preserving-type-constructors"] (NoArg guardingTypeConstructorFlag)
                     "treat type constructors as inductive constructors when checking productivity"
+    , Option []     ["no-universe-polymorphism"] (NoArg noUniversePolymorphismFlag)
+                    "disable universe polymorphism"
     , Option []     ["universe-polymorphism"] (NoArg universePolymorphismFlag)
-                    "enable universe polymorphism (experimental feature)"
+                    "enable universe polymorphism (default)"
     , Option []     ["no-irrelevant-projections"] (NoArg noIrrelevantProjectionsFlag)
                     "disable projection of irrelevant record fields"
+    , Option []     ["experimental-irrelevance"] (NoArg experimentalIrrelevanceFlag)
+                    "enable potentially unsound irrelevance features (irrelevant levels, irrelevant data matching)"
     , Option []     ["without-K"] (NoArg withoutKFlag)
                     "disable the K rule (maybe)"
     ]
@@ -405,4 +448,5 @@ usage options pluginInfos progName =
 tests :: IO Bool
 tests = runTests "Agda.Interaction.Options"
   [ quickCheck' prop_defaultOptions
+  , defaultPragmaOptionsSafe
   ]
