@@ -1,4 +1,4 @@
-{-# LANGUAGE DeriveDataTypeable, CPP #-}
+{-# LANGUAGE DeriveDataTypeable, DeriveFunctor, DeriveFoldable, DeriveTraversable, CPP #-}
 {-| The abstract syntax. This is what you get after desugaring and scope
     analysis of the concrete syntax. The type checker works on abstract syntax,
     producing internal syntax ("Agda.Syntax.Internal").
@@ -21,6 +21,7 @@ import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Info
 import Agda.Syntax.Common
+import Agda.Syntax.Fixity
 import Agda.Syntax.Position
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Literal
@@ -42,31 +43,63 @@ data Expr
 	| WithApp ExprInfo Expr [Expr]	     -- ^ with application
         | Lam  ExprInfo LamBinding Expr	     -- ^
         | AbsurdLam ExprInfo Hiding
+        | ExtendedLam ExprInfo DefInfo QName [Clause]
         | Pi   ExprInfo Telescope Expr	     -- ^
 	| Fun  ExprInfo (Arg Expr) Expr	     -- ^ independent function space
-        | Set  ExprInfo Nat		     -- ^
+        | Set  ExprInfo Nat		     -- ^ Set, Set1, Set2, ...
         | Prop ExprInfo			     -- ^
         | Let  ExprInfo [LetBinding] Expr    -- ^
         | ETel Telescope                     -- ^ only used when printing telescopes
 	| Rec  ExprInfo [(C.Name, Expr)]     -- ^ record construction
+	| RecUpdate ExprInfo Expr [(C.Name, Expr)]     -- ^ record update
 	| ScopedExpr ScopeInfo Expr	     -- ^ scope annotation
         | QuoteGoal ExprInfo Name Expr       -- ^
         | Quote ExprInfo                     -- ^
-        | DontCare                           -- ^ for printing DontCare from Syntax.Internal
+        | QuoteTerm ExprInfo                 -- ^
+        | Unquote ExprInfo                   -- ^ The splicing construct: unquote ...
+        | DontCare Expr                      -- ^ for printing DontCare from Syntax.Internal
   deriving (Typeable, Data, Show)
 
 data Declaration
 	= Axiom      DefInfo Relevance QName Expr          -- ^ postulate
 	| Field      DefInfo QName (Arg Expr)		   -- ^ record field
 	| Primitive  DefInfo QName Expr			   -- ^ primitive function
-	| Definition DeclInfo [TypeSignature] [Definition] -- ^ a bunch of mutually recursive definitions
+	| Mutual     DeclInfo [Declaration]                -- ^ a bunch of mutually recursive definitions
 	| Section    ModuleInfo ModuleName [TypedBindings] [Declaration]
-	| Apply	     ModuleInfo ModuleName [TypedBindings] ModuleName [NamedArg Expr] (Map QName QName) (Map ModuleName ModuleName)
+	| Apply	     ModuleInfo ModuleName ModuleApplication (Map QName QName) (Map ModuleName ModuleName)
 	| Import     ModuleInfo ModuleName
 	| Pragma     Range	Pragma
         | Open       ModuleInfo ModuleName
           -- ^ only retained for highlighting purposes
+        | FunDef     DefInfo QName [Clause]
+        | DataSig    DefInfo QName Telescope Expr -- ^ lone data signature
+            -- ^ the 'LamBinding's are 'DomainFree' and binds the parameters of the datatype.
+        | DataDef    DefInfo QName [LamBinding] [Constructor]
+            -- ^ the 'LamBinding's are 'DomainFree' and binds the parameters of the datatype.
+        | RecSig     DefInfo QName Telescope Expr -- ^ lone record signature
+        | RecDef     DefInfo QName (Maybe QName) [LamBinding] Expr [Declaration]
+            -- ^ The 'Expr' gives the constructor type telescope, @(x1 : A1)..(xn : An) -> Prop@,
+            --   and the optional name is the constructor's name.
 	| ScopedDecl ScopeInfo [Declaration]  -- ^ scope annotation
+        deriving (Typeable, Data, Show)
+
+class GetDefInfo a where
+  getDefInfo :: a -> Maybe DefInfo
+
+instance GetDefInfo Declaration where
+  getDefInfo (Axiom i _ _ _) = Just i
+  getDefInfo (Field i _ _) = Just i
+  getDefInfo (Primitive i _ _) = Just i
+  getDefInfo (ScopedDecl _ (d:_)) = getDefInfo d
+  getDefInfo (FunDef i _ _) = Just i
+  getDefInfo (DataSig i _ _ _) = Just i
+  getDefInfo (DataDef i _ _ _) = Just i
+  getDefInfo (RecSig i _ _ _) = Just i
+  getDefInfo (RecDef i _ _ _ _ _) = Just i
+  getDefInfo _ = Nothing
+
+data ModuleApplication = SectionApp [TypedBindings] ModuleName [NamedArg Expr]
+                       | RecordModuleIFS ModuleName
   deriving (Typeable, Data, Show)
 
 data Pragma = OptionsPragma [String]
@@ -75,38 +108,29 @@ data Pragma = OptionsPragma [String]
             | CompiledTypePragma QName String
             | CompiledDataPragma QName String [String]
             | CompiledEpicPragma QName String
+            | CompiledJSPragma QName String
+            | StaticPragma QName
             | EtaPragma QName
   deriving (Typeable, Data, Show)
 
-data LetBinding = LetBind LetInfo Relevance Name Expr Expr    -- ^ LetBind info name type defn
-                | LetApply ModuleInfo ModuleName [TypedBindings] ModuleName [NamedArg Expr] (Map QName QName) (Map ModuleName ModuleName)
+data LetBinding = LetBind LetInfo Relevance Name Expr Expr    -- ^ LetBind info rel name type defn
+                | LetApply ModuleInfo ModuleName ModuleApplication (Map QName QName) (Map ModuleName ModuleName)
                 | LetOpen ModuleInfo ModuleName     -- ^ only for highlighting and abstractToConcrete
   deriving (Typeable, Data, Show)
 
--- | A definition without its type signature.
-data Definition
-	= FunDef     DefInfo QName [Clause]
-	| DataDef    DefInfo QName [LamBinding] [Constructor]
-	    -- ^ the 'LamBinding's are 'DomainFree' and binds the parameters of the datatype.
-	| RecDef     DefInfo QName (Maybe Constructor) [LamBinding] Expr [Declaration]
-	    -- ^ The 'Expr' gives the constructor type telescope, @(x1 : A1)..(xn : An) -> Prop@,
-            --   and the optional name is the constructor's name.
-        | ScopedDef ScopeInfo Definition
-  deriving (Typeable, Data, Show)
-
 -- | Only 'Axiom's.
 type TypeSignature  = Declaration
 type Constructor    = TypeSignature
 
 -- | A lambda binding is either domain free or typed.
 data LamBinding
-	= DomainFree Hiding Name    -- ^ . @x@ or @{x}@
+	= DomainFree Hiding Relevance Name  -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@
 	| DomainFull TypedBindings  -- ^ . @(xs:e)@ or @{xs:e}@
   deriving (Typeable, Data, Show)
 
 -- | Typed bindings with hiding information.
-data TypedBindings = TypedBindings Range (Arg [TypedBinding])
-	    -- ^ . @(xs:e;..;ys:e')@ or @{xs:e;..;ys:e'}@
+data TypedBindings = TypedBindings Range (Arg TypedBinding)
+	    -- ^ . @(xs : e)@ or @{xs : e}@
   deriving (Typeable, Data, Show)
 
 -- | A typed binding. Appears in dependent function spaces, typed lambdas, and
@@ -148,7 +172,7 @@ data Pattern' e	= VarP Name
 		| AbsurdP PatInfo
 		| LitP Literal
 		| ImplicitP PatInfo	-- ^ generated at type checking for implicit arguments
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
 
 type Pattern = Pattern' Expr
 
@@ -156,48 +180,9 @@ type Pattern = Pattern' Expr
     Instances
  --------------------------------------------------------------------------}
 
-instance Functor Pattern' where
-    fmap f p = case p of
-	VarP x	    -> VarP x
-	ConP i c ps -> ConP i c $ (fmap . fmap . fmap . fmap) f ps
-	DefP i c ps -> DefP i c $ (fmap . fmap . fmap . fmap) f ps
-	LitP l	    -> LitP l
-	AsP i x p   -> AsP i x $ fmap f p
-	DotP i e    -> DotP i (f e)
-	AbsurdP i   -> AbsurdP i
-	WildP i	    -> WildP i
-	ImplicitP i -> ImplicitP i
-
--- foldr should really take its arguments in a different order!
-instance Foldable Pattern' where
-    foldr f z p = case p of
-	VarP _	    -> z
-	ConP _ _ ps -> (foldrF . foldrF . foldrF . foldrF) f ps z
-	DefP _ _ ps -> (foldrF . foldrF . foldrF . foldrF) f ps z
-	LitP _	    -> z
-	AsP _ _ p   -> foldr f z p
-	DotP _ e    -> f e z
-	AbsurdP _   -> z
-	WildP _	    -> z
-	ImplicitP _ -> z
-	where
-	    foldrF f = flip (foldr f)
-
-instance Traversable Pattern' where
-    traverse f p = case p of
-	VarP x	    -> pure $ VarP x
-	ConP i c ps -> ConP i c <$> (traverse . traverse . traverse . traverse) f ps
-	DefP i c ps -> DefP i c <$> (traverse . traverse . traverse . traverse) f ps
-	LitP l	    -> pure $ LitP l
-	AsP i x p   -> AsP i x <$> traverse f p
-	DotP i e    -> DotP i <$> f e
-	AbsurdP i   -> pure $ AbsurdP i
-	WildP i	    -> pure $ WildP i
-	ImplicitP i -> pure $ ImplicitP i
-
 instance HasRange LamBinding where
-    getRange (DomainFree _ x) = getRange x
-    getRange (DomainFull b)   = getRange b
+    getRange (DomainFree _ _ x) = getRange x
+    getRange (DomainFull b)     = getRange b
 
 instance HasRange TypedBindings where
     getRange (TypedBindings r _) = r
@@ -217,35 +202,38 @@ instance HasRange Expr where
     getRange (WithApp i _ _)	= getRange i
     getRange (Lam i _ _)	= getRange i
     getRange (AbsurdLam i _)    = getRange i
+    getRange (ExtendedLam i _ _ _)  = getRange i
     getRange (Pi i _ _)		= getRange i
     getRange (Fun i _ _)	= getRange i
     getRange (Set i _)		= getRange i
     getRange (Prop i)		= getRange i
     getRange (Let i _ _)	= getRange i
     getRange (Rec i _)		= getRange i
+    getRange (RecUpdate i _ _)  = getRange i
     getRange (ETel tel)         = getRange tel
     getRange (ScopedExpr _ e)	= getRange e
     getRange (QuoteGoal _ _ e)	= getRange e
     getRange (Quote i)  	= getRange i
-    getRange (DontCare)         = noRange
+    getRange (QuoteTerm i)  	= getRange i
+    getRange (Unquote i)  	= getRange i
+    getRange (DontCare{})       = noRange
 
 instance HasRange Declaration where
     getRange (Axiom      i _ _ _       ) = getRange i
     getRange (Field      i _ _         ) = getRange i
-    getRange (Definition i _ _	       ) = getRange i
+    getRange (Mutual     i _           ) = getRange i
     getRange (Section    i _ _ _       ) = getRange i
-    getRange (Apply	 i _ _ _ _ _ _ ) = getRange i
+    getRange (Apply	 i _ _ _ _     ) = getRange i
     getRange (Import     i _	       ) = getRange i
     getRange (Primitive  i _ _	       ) = getRange i
     getRange (Pragma	 i _	       ) = getRange i
     getRange (Open       i _           ) = getRange i
     getRange (ScopedDecl _ d	       ) = getRange d
-
-instance HasRange Definition where
-    getRange (FunDef  i _ _    )   = getRange i
-    getRange (DataDef i _ _ _  )   = getRange i
-    getRange (RecDef  i _ _ _ _ _) = getRange i
-    getRange (ScopedDef _ d)       = getRange d
+    getRange (FunDef  i _ _         ) = getRange i
+    getRange (DataSig i _ _ _       ) = getRange i
+    getRange (DataDef i _ _ _       ) = getRange i
+    getRange (RecSig  i _ _ _       ) = getRange i
+    getRange (RecDef  i _ _ _ _ _   ) = getRange i
 
 instance HasRange (Pattern' e) where
     getRange (VarP x)	   = getRange x
@@ -272,12 +260,12 @@ instance HasRange RHS where
 
 instance HasRange LetBinding where
     getRange (LetBind  i _ _ _ _     ) = getRange i
-    getRange (LetApply i _ _ _ _ _ _ ) = getRange i
+    getRange (LetApply i _ _ _ _     ) = getRange i
     getRange (LetOpen  i _           ) = getRange i
 
 instance KillRange LamBinding where
-  killRange (DomainFree h x) = killRange1 (DomainFree h) x
-  killRange (DomainFull b)   = killRange1 DomainFull b
+  killRange (DomainFree h r x) = killRange1 (DomainFree h r) x
+  killRange (DomainFull b)     = killRange1 DomainFull b
 
 instance KillRange TypedBindings where
   killRange (TypedBindings r b) = TypedBindings (killRange r) (killRange b)
@@ -297,17 +285,21 @@ instance KillRange Expr where
   killRange (WithApp i e es) = killRange3 WithApp i e es
   killRange (Lam i b e)      = killRange3 Lam i b e
   killRange (AbsurdLam i h)  = killRange1 AbsurdLam i h
+  killRange (ExtendedLam i name di pes) = killRange4 ExtendedLam i name di pes --(\_ -> ExtendedLam i def {-name di si-}) i pes
   killRange (Pi i a b)       = killRange3 Pi i a b
   killRange (Fun i a b)      = killRange3 Fun i a b
   killRange (Set i n)        = Set (killRange i) n
   killRange (Prop i)         = killRange1 Prop i
   killRange (Let i ds e)     = killRange3 Let i ds e
   killRange (Rec i fs)       = Rec (killRange i) (map (id -*- killRange) fs)
+  killRange (RecUpdate i e fs) = RecUpdate (killRange i) (killRange e) (map (id -*- killRange) fs)
   killRange (ETel tel)       = killRange1 ETel tel
   killRange (ScopedExpr s e) = killRange1 (ScopedExpr s) e
   killRange (QuoteGoal i x e)= killRange3 QuoteGoal i x e
   killRange (Quote i)        = killRange1 Quote i
-  killRange (DontCare)       = DontCare
+  killRange (QuoteTerm i)    = killRange1 QuoteTerm i
+  killRange (Unquote i)      = killRange1 Unquote i
+  killRange (DontCare e)     = DontCare e
 
 instance KillRange Relevance where
   killRange rel = rel -- no range to kill
@@ -315,21 +307,27 @@ instance KillRange Relevance where
 instance KillRange Declaration where
   killRange (Axiom      i rel a b     ) = killRange4 Axiom      i rel a b
   killRange (Field      i a b         ) = killRange3 Field      i a b
-  killRange (Definition i a b         ) = killRange3 Definition i a b
+  killRange (Mutual     i a           ) = killRange2 Mutual     i a
   killRange (Section    i a b c       ) = killRange4 Section    i a b c
-  killRange (Apply      i a b c d e f ) = killRange5 Apply      i a b c d e f
+  killRange (Apply      i a b c d     ) = killRange3 Apply      i a b c d
    -- the last two arguments of Apply are name maps, so nothing to kill
   killRange (Import     i a           ) = killRange2 Import     i a
   killRange (Primitive  i a b         ) = killRange3 Primitive  i a b
   killRange (Pragma     i a           ) = Pragma (killRange i) a
   killRange (Open       i x           ) = killRange2 Open       i x
   killRange (ScopedDecl a d           ) = killRange1 (ScopedDecl a) d
+  killRange (FunDef  i a b       ) = killRange3 FunDef  i a b
+  killRange (DataSig i a b c     ) = killRange3 DataSig i a b c
+  killRange (DataDef i a b c     ) = killRange4 DataDef i a b c
+  killRange (RecSig  i a b c     ) = killRange4 RecSig  i a b c
+  killRange (RecDef  i a b c d e ) = killRange6 RecDef  i a b c d e
 
-instance KillRange Definition where
-  killRange (FunDef  i a b    )   = killRange3 FunDef  i a b
-  killRange (DataDef i a b c)     = killRange4 DataDef i a b c
-  killRange (RecDef  i a b c d e) = killRange6 RecDef  i a b c d e
-  killRange (ScopedDef s a)       = killRange1 (ScopedDef s) a
+instance KillRange ModuleApplication where
+  killRange (SectionApp a b c  ) = killRange3 SectionApp a b c
+  killRange (RecordModuleIFS a ) = killRange1 RecordModuleIFS a
+
+instance KillRange x => KillRange (ThingWithFixity x) where
+  killRange (ThingWithFixity c f) = ThingWithFixity (killRange c) f
 
 instance KillRange e => KillRange (Pattern' e) where
   killRange (VarP x)      = killRange1 VarP x
@@ -356,7 +354,7 @@ instance KillRange RHS where
 
 instance KillRange LetBinding where
   killRange (LetBind  i rel a b c   ) = killRange5 LetBind  i rel a b c
-  killRange (LetApply i a b c d e f ) = killRange5 LetApply i a b c d e f
+  killRange (LetApply i a b c d     ) = killRange3 LetApply i a b c d
   killRange (LetOpen  i x           ) = killRange2 LetOpen  i x
 
 ------------------------------------------------------------------------
@@ -365,32 +363,80 @@ instance KillRange LetBinding where
 
 -- | Extracts all the names which are declared in a 'Declaration'.
 -- This does not include open public or let expressions, but it does
--- include local modules and where clauses.
+-- include local modules, where clauses and the names of extended
+-- lambdas.
 
 allNames :: Declaration -> Seq QName
-allNames (Axiom     _ _ q _)   = Seq.singleton q
-allNames (Field     _   q _)   = Seq.singleton q
-allNames (Primitive _   q _)   = Seq.singleton q
-allNames (Definition _ _ defs) = Fold.foldMap allNamesD defs
+allNames (Axiom     _ _ q _)      = Seq.singleton q
+allNames (Field     _   q _)      = Seq.singleton q
+allNames (Primitive _   q _)      = Seq.singleton q
+allNames (Mutual     _ defs)      = Fold.foldMap allNames defs
+allNames (DataSig _ q _ _)        = Seq.singleton q
+allNames (DataDef _ q _ decls)    = q <| Fold.foldMap allNames decls
+allNames (RecSig _ q _ _)         = Seq.singleton q
+allNames (RecDef _ q c _ _ decls) =
+  q <| foldMap Seq.singleton c >< Fold.foldMap allNames decls
+allNames (FunDef _ q cls)         = q <| Fold.foldMap allNamesC cls
   where
-  allNamesD :: Definition -> Seq QName
-  allNamesD (FunDef _ q cls)         = q <| Fold.foldMap allNamesC cls
-  allNamesD (DataDef _ q _ decls)    = q <| Fold.foldMap allNames decls
-  allNamesD (ScopedDef _ def)        = allNamesD def
-  allNamesD (RecDef _ q c _ _ decls) =
-    q <| Fold.foldMap allNames c >< Fold.foldMap allNames decls
-
   allNamesC :: Clause -> Seq QName
   allNamesC (Clause _ rhs decls) = allNamesR rhs ><
                                    Fold.foldMap allNames decls
 
   allNamesR :: RHS -> Seq QName
-  allNamesR RHS {}                = Seq.empty
+  allNamesR (RHS e)               = allNamesE e
   allNamesR AbsurdRHS {}          = Seq.empty
   allNamesR (WithRHS q _ cls)     = q <| Fold.foldMap allNamesC cls
   allNamesR (RewriteRHS qs _ rhs cls) =
     Seq.fromList qs >< allNamesR rhs
                     >< Fold.foldMap allNames cls
+
+  allNamesE :: Expr -> Seq QName
+  allNamesE Var {}                  = Seq.empty
+  allNamesE Def {}                  = Seq.empty
+  allNamesE Con {}                  = Seq.empty
+  allNamesE Lit {}                  = Seq.empty
+  allNamesE QuestionMark {}         = Seq.empty
+  allNamesE Underscore {}           = Seq.empty
+  allNamesE (App _ e1 e2)           = Fold.foldMap allNamesE [e1, namedThing (unArg e2)]
+  allNamesE (WithApp _ e es)        = Fold.foldMap allNamesE (e : es)
+  allNamesE (Lam _ b e)             = allNamesLam b >< allNamesE e
+  allNamesE AbsurdLam {}            = Seq.empty
+  allNamesE (ExtendedLam _ _ q cls) = q <| Fold.foldMap allNamesC cls
+  allNamesE (Pi _ tel e)            = Fold.foldMap allNamesBinds tel ><
+                                                allNamesE e
+  allNamesE (Fun _ (Arg _ _ e1) e2) = Fold.foldMap allNamesE [e1, e2]
+  allNamesE Set {}                  = Seq.empty
+  allNamesE Prop {}                 = Seq.empty
+  allNamesE (Let _ lbs e)           = Fold.foldMap allNamesLet lbs ><
+                                                allNamesE e
+  allNamesE ETel {}                 = __IMPOSSIBLE__
+  allNamesE (Rec _ fields)          = Fold.foldMap allNamesE (map snd fields)
+  allNamesE (RecUpdate _ e fs)      = allNamesE e >< Fold.foldMap allNamesE (map snd fs)
+  allNamesE (ScopedExpr _ e)        = allNamesE e
+  allNamesE (QuoteGoal _ _ e)       = allNamesE e
+  allNamesE Quote {}                = Seq.empty
+  allNamesE QuoteTerm {}            = Seq.empty
+  allNamesE Unquote {}              = Seq.empty
+  allNamesE DontCare {}             = Seq.empty
+
+  allNamesLam :: LamBinding -> Seq QName
+  allNamesLam DomainFree {}      = Seq.empty
+  allNamesLam (DomainFull binds) = allNamesBinds binds
+
+  allNamesBinds :: TypedBindings -> Seq QName
+  allNamesBinds (TypedBindings _ (Arg _ _ (TBind _ _ e))) = allNamesE e
+  allNamesBinds (TypedBindings _ (Arg _ _ (TNoBind e)))   = allNamesE e
+
+  allNamesLet :: LetBinding -> Seq QName
+  allNamesLet (LetBind _ _ _ e1 e2)  = Fold.foldMap allNamesE [e1, e2]
+  allNamesLet (LetApply _ _ app _ _) = allNamesApp app
+  allNamesLet LetOpen {}             = Seq.empty
+
+  allNamesApp :: ModuleApplication -> Seq QName
+  allNamesApp (SectionApp bindss _ es) = Fold.foldMap allNamesBinds bindss ><
+                                         Fold.foldMap allNamesE (map (namedThing . unArg) es)
+  allNamesApp RecordModuleIFS {}       = Seq.empty
+
 allNames (Section _ _ _ decls) = Fold.foldMap allNames decls
 allNames Apply {}              = Seq.empty
 allNames Import {}             = Seq.empty
@@ -405,3 +451,26 @@ allNames (ScopedDecl _ decls)  = Fold.foldMap allNames decls
 axiomName :: Declaration -> QName
 axiomName (Axiom _ _ q _) = q
 axiomName _             = __IMPOSSIBLE__
+
+-- | Are we in an abstract block?
+--
+--   In that case some definition is abstract.
+class AnyAbstract a where
+  anyAbstract :: a -> Bool
+
+instance AnyAbstract a => AnyAbstract [a] where
+  anyAbstract = Fold.any anyAbstract
+
+instance AnyAbstract Declaration where
+  anyAbstract (Axiom i _ _ _)     = defAbstract i == AbstractDef
+  anyAbstract (Field i _ _)       = defAbstract i == AbstractDef
+  anyAbstract (Mutual     _ ds)   = anyAbstract ds
+  anyAbstract (ScopedDecl _ ds)   = anyAbstract ds
+  anyAbstract (Section _ _ _ ds)  = anyAbstract ds
+  anyAbstract (FunDef i _ _)       = defAbstract i == AbstractDef
+  anyAbstract (DataDef i _ _ _)    = defAbstract i == AbstractDef
+  anyAbstract (RecDef i _ _ _ _ _) = defAbstract i == AbstractDef
+  anyAbstract (DataSig i _ _ _)    = defAbstract i == AbstractDef
+  anyAbstract (RecSig i _ _ _)     = defAbstract i == AbstractDef
+  anyAbstract _                    = __IMPOSSIBLE__
+
