@@ -27,7 +27,7 @@ getArity x = do
   def <- theDef <$> getConstInfo x
   case def of
     Axiom{}                                 -> return 0
-    Function{ funClauses = c : _ }          -> return $ genericLength (clausePats $ translatedClause c)
+    Function{ funClauses = c : _ }          -> return $ genericLength (clausePats c)
     Function{ funClauses = [] }             -> return 0
     Datatype{ dataPars = np, dataIxs = ni } -> return np
     Record{ recPars = n }                   -> return n
@@ -136,7 +136,8 @@ instance HasPolarity a => HasPolarity (Arg a) where
   polarities i = polarities i . unArg
 
 instance HasPolarity a => HasPolarity (Abs a) where
-  polarities i = polarities (i + 1) . absBody
+  polarities i (Abs   _ b) = polarities (i + 1) b
+  polarities i (NoAbs _ v) = polarities i v
 
 instance HasPolarity a => HasPolarity [a] where
   polarities i xs = concat <$> mapM (polarities i) xs
@@ -153,13 +154,27 @@ instance HasPolarity Term where
               | otherwise -> polarities i ts
     Lam _ t    -> polarities i t
     Lit _      -> return []
+    Level l    -> polarities i l
     Def x ts   -> do
       pols <- getPolarity x
       let compose p ps = map (composePol p) ps
       concat . zipWith compose (pols ++ repeat Invariant) <$> mapM (polarities i) ts
     Con _ ts   -> polarities i ts
     Pi a b     -> (++) <$> (map neg <$> polarities i a) <*> polarities i b
-    Fun a b    -> (++) <$> (map neg <$> polarities i a) <*> polarities i b
     Sort _     -> return []
     MetaV _ ts -> map (const Invariant) <$> polarities i ts
-    DontCare   -> return []
+    DontCare _ -> return []
+
+instance HasPolarity Level where
+  polarities i (Max as) = polarities i as
+
+instance HasPolarity PlusLevel where
+  polarities i ClosedLevel{} = return []
+  polarities i (Plus _ l) = polarities i l
+
+instance HasPolarity LevelAtom where
+  polarities i l = case l of
+    MetaLevel _ vs   -> map (const Invariant) <$> polarities i l
+    BlockedLevel _ v -> polarities i v
+    NeutralLevel v   -> polarities i v
+    UnreducedLevel v -> polarities i v
